fizz = {}
fizz.__index = fizz
fizz.world = {}
fizz.world.__index = fizz.world
fizz.body = {}
fizz.body.__index = fizz.body
fizz.body.list = {}
fizz.shape = {}
fizz.shape.__index = fizz.shape
fizz.shape.list = {}
fizz.joint = {}
fizz.joint.__index = fizz.joint
fizz.collision = {}
fizz.collision.__index = fizz.collision
fizz.collisions = {}

--Draw Settings
fizz.drawLines = true
fizz.drawFill = true
fizz.circleQuality = 20
fizz.boundLineWidth = 1
fizz.boundPadding = 10
fizz.boundColor = {255,20,50}
fizz.bodySleepingLineColor = {229, 127, 127}
fizz.bodySleepingFillColor = {229, 127, 127, 255/8}
fizz.bodyStaticLineColor = {127, 229, 127}
fizz.bodyStaticFillColor = {127, 229, 127, 255/8}
fizz.shapeSensorLineColor = { 90, 90, 90 }
fizz.shapeSensorFillColor = { 90, 90, 90, 255/8 }
fizz.shapeHoverLineColor = {255, 120, 0}
fizz.shapeHoverFillColor = {255, 120, 0, 255/8}
fizz.shapeLineColor = {255, 255, 255}
fizz.shapeFillColor = {255, 255, 255, 255/8}
fizz.jointLineColor = {255, 255, 0}
fizz.revoluteJointColor = fizz.jointLineColor
fizz.font = love.graphics.newFont("gfx/fizzDebug.ttf", 10)

--WORLD CREATION
function fizz.newWorld(w, h)
	local o = {}
	o.world = love.physics.newWorld(w, h)
	o.world:setCallbacks(fizz.collisionAdd, nil, fizz.collisionRemove, nil)
	
	setmetatable(o, fizz.world)
	
	return o
end

--BODY CREATION
function fizz.world:newBody(x, y, m, i)
	local o = {}
	
	o.world = self.world
	o.body = love.physics.newBody(o.world, x, y, m, i)
	o.shapes = {}
	o.joints = {}
	o.collision = false
	
	o.debugDraw = true
	
	setmetatable(o, fizz.body)
	table.insert(fizz.body.list, o)
	
	return o
end

--SHAPE CREATION
function fizz.body:newCircleShape(xo, yo, r)
	local o = {}
	o.shape = love.physics.newCircleShape(self.body, xo, yo, r)
	o.xo = xo or 0
	o.yo = yo or 0
	o.r = r or 0
	o.collision = false
	o.shape:setData({body=self, shape=o.shape})
	
	setmetatable(o, fizz.shape)
	table.insert(self.shapes, o)
	
	print("Creating Circle Shape " .. tostring(o) )
	return o
end

function fizz.body:newRectangleShape(xo, yo, w, h, a)
	local o = {}
	o.shape = love.physics.newRectangleShape(self.body, xo, yo, w, h, a)
	o.xo = xo or 0
	o.yo = yo or 0
	o.w = w or 0
	o.h = h or 0
	o.a = a or 0
	o.collision = false
	o.shape:setData({body=self, shape=o.shape})
		
	setmetatable(o, fizz.shape)
	table.insert(self.shapes, o)
	
	print("Creating Rectangle Shape: " .. tostring(o) )
	return o
end

function fizz.body:newPolygonShape(...)
	local o = {}
	o.shape = love.physics.newPolygonShape(self.body, ...)
	o.collision = false
	o.shape:setData({body=self, shape=o.shape})
	
	setmetatable(o, fizz.shape)
	table.insert(self.shapes, o)
	
	print("Creating Polygon Shape: " .. tostring(o) )
	return o
end

	--JOINT CREATION
function fizz.body:newDistanceJoint(oBody, xo1, yo1, xo2, yo2)
	local j = {}
	
	j.b1 = self
	j.b2 = oBody
	j.xo1 = xo1 or 0
	j.xo2 = xo2 or 0
	j.yo1 = yo1 or 0
	j.yo2 = yo2 or 0
	j.joint = love.physics.newDistanceJoint(j.b1.body, j.b2.body, j.b1.body:getX()+j.xo1, j.b1.body:getY()+j.yo1, j.b2.body:getX()+j.xo2, j.b2.body:getY()+j.yo2 )
	j.jointtype = "distance"
	
	setmetatable(j, fizz.joint)
	table.insert(self.joints, j)

	return j
end

function fizz.body:newGearJoint(j1, j2, ratio)
	local j = {}
	j.j1 = j1.joint
	j.j2 = j2.joint
	j.ratio = ratio
	j.joint = love.physics.newGearJoint( j.j1, j.j2, j.ratio )
	j.jointtype = "gear"
	
	setmetatable(j, fizz.joint)
	table.insert(self.joints, j)
	
	return j
end
	
function fizz.body:newMouseJoint(x, y)
	local j = {}
	j.b1 = self
	j.x = x
	j.y = y
	j.joint = love.physics.newMouseJoint(j.b1.body, j.x, j.y)
	j.jointtype = "mouse"
	
	setmetatable(j, fizz.joint)
	table.insert(self.joints, j)
	
	return j
end

function fizz.body:newPrismaticJoint(oBody, x, y, ax, ay)
	local j = {}
	j.b1 = self
	j.b2 = oBody
	j.x = x or 0
	j.y = y or 0
	j.ax = ax
	j.ay = ay
	j.joint = love.physics.newPrismaticJoint(j.b1.body, j.b2.body, j.b1.body:getX()+j.x, j.b1.body:getY()+j.y, j.ax, j.ay)
	j.jointtype = "prismatic"
	
	setmetatable(j, fizz.joint)
	table.insert(self.joints, j)
	table.insert(oBody.joints, j)
	
	return j	
end

function fizz.body:newPulleyJoint(oBody, gx1, gy1, gx2, gy2, x1, y1, x2, y2, ratio)
	local j = {}
	j.b1 = self
	j.b2 = oBody
	j.gx1 = gx1
	j.gy1 = gy1
	j.gx2 = gx2
	j.gy2 = gy2
	j.x1 = x1
	j.y1 = y1
	j.x2 = x2
	j.y2 = y2
	j.ratio = ratio
	j.joint = love.physics.newPulleyJoint(j.b1, j.b2, j.gx1, j.gy1, j.gx2, j.gy2, j.x1, j.y1, j.x2, j.y2, j.ratio )
	j.jointtype = "pulley"
	
	setmetatable(j, fizz.joint)
	table.insert(self.joints, j)
	table.insert(oBody.joints, j)
	
	return j
end

function fizz.body:newRevoluteJoint(oBody, xo, yo)
	local j = {}
	
	j.b1 = self
	j.b2 = oBody
	j.xo = xo or 0
	j.yo = yo or 0
	j.joint = love.physics.newRevoluteJoint(j.b1.body, j.b2.body, self.body:getX()+j.xo, self.body:getY()+j.yo)
	j.jointtype = "revolute"

	setmetatable(j, fizz.joint)
	table.insert(self.joints, j)
	table.insert(oBody.joints, j)
	
	return j
end

--WORLD METHODS
function fizz.world:getBodyCount()
	return self.world:getBodyCount()
end

function fizz.world:getCallbacks()
	return self.world:getCallbacks()
end

function fizz.world:getGravity()
	return self.world:getGravity()
end

function fizz.world:getJointCount()
	return self.world:getJointCount()
end

function fizz.world:getMeter()
	return self.world:getMeter()
end

function fizz.world:isAllowSleep()
	return self.world:isAllowSleep()
end

function fizz.world:setAllowSleep(bool)
	self.world:setAllowSleep(bool)
end

function fizz.world:setCallbacks(...)
	self.world:setCallbacks(...)
end

function fizz.world:setGravity(xg, yg)
	self.world:setGravity(xg, yg)
end

function fizz.world:setMeter(scale)
	self.world:setMeter(scale)
end

function fizz.world:update(dt)
	self.world:update(dt)
end

--BODY METHODS
function fizz.body:applyForce(x, y)
	self.body:applyForce(x, y)
end

function fizz.body:applyImpulse(x, y)
	self.body:applyImpulse(x, y)
end

function fizz.body:applyTorque(torque)
	self.body:applyTorque(torque)
end

function fizz.body:destroy()
	return self.body:destroy()
end

function fizz.body:getAngle()
	return self.body:getAngle()
end

function fizz.body:getAngularDamping()
	return self.body:getAngularDamping()
end

function fizz.body:getAngularVelocity()
	return self.body:getAngularVelocity()
end

function fizz.body:getInertia()
	return self.body:getInertia()
end

function fizz.body:getLinearDamping()
	return self.body:getLinearDamping()
end

function fizz.body:getLinearVelocity()
	return self.body:getLinearVelocity()
end

function fizz.body:getLinearVelocityFromLocalPoint()
	return self.body:getLinearVelocityFromLocalPoint()
end

function fizz.body:getLinearVelocityFromWorldPoint()
	return self.body:getLinearVelocityFromWorldPoint()
end

function fizz.body:getLocalCenter()
	return self.body:getLocalCenter()
end

function fizz.body:getLocalPoint()
	return self.body:getLocalPoint()
end

function fizz.body:getLocalVector()
	return self.body:getLocalVector()
end

function fizz.body:getMass()
	return self.body:getMass()
end

function fizz.body:getPosition()
	return self.body:getPosition()
end

function fizz.body:getWorldCenter()
	return self.body:getWorldCenter()
end

function fizz.body:getWorldPoint()
	return self.body:getWorldPoint()
end

function fizz.body:getWorldVector()
	return self.body:getWorldVector()
end

function fizz.body:getX()
	return self.body:getX()
end

function fizz.body:getY()
	return self.body:getY()
end

function fizz.body:isBullet()
	return self.body:isBullet()
end

function fizz.body:isDynamic()
	return self.body:isDynamic()
end

function fizz.body:isFrozen()
	return self.body:isFrozen()
end

function fizz.body:isSleeping()
	return self.body:isSleeping()
end

function fizz.body:isStatic()
	return self.body:isStatic()
end

function fizz.body:putToSleep()
	self.body:putToSleep()
end

function fizz.body:setAllowSleeping(bool)
	self.body:setAllowSleeping(bool)
end

function fizz.body:setAngle(angle)
	self.body:setAngle(angle)
end

function fizz.body:setAngularDamping(damping)
	self.body:setAngularDamping(damping)
end

function fizz.body:setAngularVelocity(velocity)
	self.body:setAngularVelocity(velocity)
end

function fizz.body:setBullet(bool)
	self.body:setBullet(bool)
end

function fizz.body:setFixedRotation(bool)
	self.body:setFixedRotation(bool)
end

function fizz.body:setInertia(inertia)
	self.body:setInertia(inertia)
end

function fizz.body:setLinearDamping(damping)
	self.body:setLinearDamping(damping)
end

function fizz.body:setLinearVelocity(xVelocity, yVelocity)
	self.body:setLinearVelocity(xVelocity, yVelocity)
end

function fizz.body:setMass(mass)
	self.body:setMass(mass)
end

function fizz.body:setMassFromShapes()
	self.body:setMassFromShapes()
end

function fizz.body:setPosition(x, y)
	self.body:setPosition(x, y)
end

function fizz.body:setX(x)
	self.body:setX(x)
end

function fizz.body:setY(y)
	self.body:setY(y)
end

function fizz.body:wakeUp()
	self.body:wakeUp()
end

function fizz.body:setDebugDraw(bool)
	self.debugDraw = bool
end

function fizz.body:getDebugDraw()
	return self.debugDraw
end

--SHAPE DEFINITIONS
function fizz.shape:destroy()
	self.shape:destroy()
	self.joint = nil
	self = nil
end

function fizz.shape:getBody()
	return self.shape:getBody()
end

function fizz.shape:getBoundingBox()
	return self.shape:getBoundingBox()
end

function fizz.shape:getCategory()
	return self.shape:getCategory()
end

function fizz.shape:getCategoryBits()
	return self.shape:getCategoryBits()
end

function fizz.shape:getData()
	return self.shape:getData()
end

function fizz.shape:getDensity()
	return self.shape:getDensity()
end

function fizz.shape:getFilterData()
	return self.shape:getFilterData()
end

function fizz.shape:getFriction()
	return self.shape:getFriction()
end

function fizz.shape:getMask()
	return self.shape:getMask()
end

function fizz.shape:getMaskBits()
	return self.shape:getMaskBits()
end

function fizz.shape:getRestitution()
	return self.shape:getRestitution()
end

function fizz.shape:getType()
	return self.shape:getType()
end

function fizz.shape:isSensor()
	return self.shape:isSensor()
end

function fizz.shape:setCategory(...)
	self.shape:setCategory(...)
end

function fizz.shape:setData(data)
	self.shape:setData(data)
end

function fizz.shape:setDensity(density)
	self.shape:setDensity(density)
end

function fizz.shape:setFilterData(categoryBits, maskBits, groupIndex)
	self.shape:setFilterData(categoryBits, maskBits, groupIndex)
end

function fizz.shape:setFriction(friction)
	self.shape:setFriction(friction)
end

function fizz.shape:setMask(...)
	self.shape:setMask(...)
end

function fizz.shape:setRestitution(restitution)
	self.shape:setRestitution(restitution)
end

function fizz.shape:setSensor(sensor)
	self.shape:setSensor(sensor)
end

function fizz.shape:testPoint(x, y)
	return self.shape:testPoint(x, y)
end

function fizz.shape:testSegment(x1, y1, x2, y2)
	return self.shape:testSegment(x1, y1, x2, y2)
end

function fizz.shape:getPoints() --Polygon & Rectangle shapes only
	return self.shape:getPoints()
end

function fizz.shape:getLocalCenter() --Circle shapes only
	return self.shape:getLocalCenter()
end

function fizz.shape:getRadius() --Circle shapes only
	return self.shape:getRadius()
end

function fizz.shape:getWorldCenter()
	return self.shape:getWorldCenter()
end

--JOINT METHODS
function fizz.joint:update() -- Would be used for dynamic joints (e.g - Springs, or Wobble Revolute joints)
end

function fizz.joint:destroy()
	self.joint:destroy()
	self.joint = nil
	self = nil
end

--[[

	!!COLLISIONS!! :(

]]--
function fizz.collisionAdd(body1, body2, collision)
	if body1 and body2 and collision and (not body1.body.collision and not body2.body.collision ) then
		--body1.collision = { body1=body1, body2=body2, contact=collision }
		--body2.collision = { body1=body2, body2=body1, contact=collision }
		body1.body.collision = { shape=body1.shape, contact=collision }
		body2.body.collision = { shape=body1.shape, contact=collision }
		print("Added Collision: " .. tostring(body1.shape:getType() ) .. " -> " .. tostring(body2.shape:getType())  )
		--print("----")
	end
end

function fizz.collisionRemove(body1, body2, collision)
	if body1 and body2 and collision then
		--body1.collision = false
		--body2.collision = false
		body1.body.collision = false
		body2.body.collision = false
		print("Removed collision")
	end
end

function fizz.body:drawCollisions()
	if self.collision then
		if self.collision.contact then
			local cx, cy = self.collision.contact:getPosition()
			love.graphics.circle("line", cx, cy, 8)
		end
	end
end
--[[

	!!END COLLISIONS!! :(

]]--

--MISC. FIZZ FUNCTIONS
function fizz.bodyAtPoint(x, y) --Return body at X/Y
	for i, bodyC in ipairs(fizz.body.list) do
		for ii, shape in ipairs(bodyC.shapes) do
			if shape:testPoint(x, y) then
				return bodyC
			end
		end
	end
	return false
end

function fizz.drawAll(drawShapes, drawJoints, drawCollisions) 
	fizz.bodyCount = #fizz.body.list
	for i, bodyC in ipairs(fizz.body.list) do
		if bodyC:getDebugDraw() then
			if drawShapes then
				bodyC:drawShapes()				
			end
			if drawJoints then
				bodyC:drawJoints()
			end
			if drawCollisions then
				bodyC:drawCollisions()
			end
		end
	end
end

function fizz.body:drawShapes()
	local mx, my = love.mouse.getPosition()
	local bx, by = self.body:getPosition()
	local fillColor = fizz.shapeFillColor
	local lineColor = fizz.shapeLineColor
	
	love.graphics.setColor( unpack(fizz.shapeLineColor) )
	
	if #self.shapes == 0 then
		love.graphics.line(bx-3, by-3, bx+3, by+3)
		love.graphics.line(bx-3, by+3, bx+3, by-3)
	end
	
	for i, shape in ipairs(self.shapes) do
	
		if shape:testPoint((mx-camX)/camScale, (my-camY)/camScale) then
			fillColor = fizz.shapeHoverFillColor
			lineColor = fizz.shapeHoverLineColor
		elseif self.body:isSleeping() then
			lineColor = fizz.bodySleepingLineColor
			fillColor = fizz.bodySleepingFillColor
		elseif self.body:isStatic() then
			lineColor = fizz.bodyStaticLineColor
			fillColor = fizz.bodyStaticFillColor
		elseif shape:isSensor() then
			lineColor = fizz.shapeSensorLineColor
			fillColor = fizz.shapeSensorFillColor
		else
			lineColor = fizz.shapeLineColor
			fillColor = fizz.shapeFillColor
		end
		
		if shape:getType() == "circle" then
			local lx, ly = shape:getWorldCenter( )
			local rx = lengthdir_x(shape:getRadius(), -math.deg( self.body:getAngle() ) )
			local ry = lengthdir_y(shape:getRadius(), -math.deg( self.body:getAngle() ) )
			
			if fizz.drawLines then
				love.graphics.setColor( unpack( lineColor ) )
				love.graphics.line(lx, ly, lx+rx, ly+ry)
				love.graphics.circle("line", lx, ly, shape:getRadius(), fizz.circleQuality )
			end
			
			if fizz.drawFill then
				love.graphics.setColor( unpack( fillColor ) )
				love.graphics.circle("fill", lx, ly, shape:getRadius(), fizz.circleQuality )
			end
			
		elseif shape:getType() == "polygon" then
			if fizz.drawLines then
				love.graphics.setColor( unpack( lineColor ) )
				love.graphics.polygon("line", shape:getPoints() )
			end
			
			if fizz.drawFill then
				love.graphics.setColor( unpack( fillColor ) )
				love.graphics.polygon("fill", shape:getPoints() )
			end
		end
	end
	love.graphics.setColor(255,255,255)
end

function fizz.body:drawJoints()
	for i, joint in ipairs(self.joints) do
		if joint.joint then
			love.graphics.setColor(unpack(fizz.jointLineColor))
			local x1, y1, x2, y2 = joint.joint:getAnchors()
			love.graphics.line(x1, y1, x2, y2)
			if joint.jointtype == "distance" or joint.jointtype == "prismatic" then
				--love.graphics.line(x1, y1, x2, y2)
			elseif joint.jointtype == "revolute" then
				love.graphics.line(joint.b1:getX(), joint.b1:getY(), x1, y1)
				love.graphics.line(joint.b2:getX(), joint.b2:getY(), x2, y2)
				
				--love.graphics.setColor(unpack(fizz.revoluteJointColor))
				--love.graphics.circle("fill", x2, y2, 2/camScale)
				--love.graphics.setLineStipple()
			end
		end
	end
	love.graphics.setColor(255,255,255)
end

function fizz.body:drawBound(drawObjectInfo)
	local bl, br, bt, bb, pbl, pbr, pbt, pbb
	bl = 2^1024
	br = -2^1024
	bt = 2^1024
	bb = -2^1024
	
	pbl = bl
	pbr = br
	pbt = bt
	pbb = bb
	
	--Calculate the four corners boundary
	for i, shape in ipairs(self.shapes) do
		local x1, y1, x2, y2, x3, y3, x4, y4 = shape:getBoundingBox( )
			
		bl = math.min(x1, x2, x3, x4, pbl)
		br = math.max(x1, x2, x3, x4, pbr)
		bt = math.min(y1, y2, y3, y4, pbt)
		bb = math.max(y1, y2, y3, y4, pbb)
		
		pbl = bl
		pbr = br
		pbt = bt
		pbb = bb
	end
	
	--Remember previous draw settings
	preLineWidth= love.graphics.getLineWidth()
	preColor = {love.graphics.getColor()}
	preFont = love.graphics.getFont()
	
	--Set boundary draw settings
	love.graphics.setLineWidth(fizz.boundLineWidth)
	love.graphics.setColor(unpack(fizz.boundColor))
	
	--Draw Boundary (With Padding)
	love.graphics.rectangle("line", bl-fizz.boundPadding, bt-fizz.boundPadding, (br-bl)+fizz.boundPadding*2, (bb-bt)+fizz.boundPadding*2)
	
	--Draw Object Information
	if drawObjectInfo then
		love.graphics.setFont(fizz.font)
		love.graphics.setColor(255,255,255)
		objectInfo = ""
		objectInfo = objectInfo .. "X: " .. string.format("%.2f", self.body:getX() ) .. "  Y: " .. string.format("%.2f", self.body:getY() ) .. "\n"
		objectInfo = objectInfo .. "ANGLE(Rad): " .. string.format("%.2f", self.body:getAngle() ) .. "\n"
		objectInfo = objectInfo .. "ANGLE(Deg): " .. string.format("%.2f", math.deg(self.body:getAngle() ) ) .. "\n"
		objectInfo = objectInfo .. "MASS: " .. string.format("%.6f", self.body:getMass() ) .. "\n"
		objectInfo = objectInfo .. "INERTIA: " .. string.format("%.6f", self.body:getInertia() ) .. "\n"
		objectInfo = objectInfo .. "Sleeping/Allowed: " .. tostring(self.body:isSleeping()) .. "\n"
		objectInfo = objectInfo .. "Dynamic: " .. tostring(self.body:isDynamic()) .. "\n"
		objectInfo = objectInfo .. "#Shapes: " .. #self.shapes .. "\n"
		objectInfo = objectInfo .. "#Joints: " .. #self.joints .. "\n"
		love.graphics.print(objectInfo, math.ceil(-camX), math.ceil(-camY + fizz.font:getHeight(objectInfo)) )
	end
	
	--Reset to previous draw settings
	love.graphics.setLineWidth(preLineWidth)
	love.graphics.setColor(unpack(preColor))
	love.graphics.setFont(preFont)
end

function lengthdir_x(len, dir)
	return math.cos(dir * math.pi / 180) * len;
end

function lengthdir_y(len, dir)
	return -math.sin(dir * math.pi / 180) * len
end

function point_distance( x1, y1, x2, y2 )
	local xd = x2 - x1
	local yd = y2 - y1
	return math.sqrt( xd * xd + yd * yd )
end