--Garage [Class for vehicles]
Garage = {}
Garage.__index = Garage
Garage.Vehicle = {}
Garage.Vehicle.__index = Garage.Vehicle

function Garage.newVehicle(x, y, vehicleFile)
	local o = {}
	
	o.x = x or 0
	o.y = y or 0
	
	local chassisPoly1 = {
		0, 49,
		13, 25,
		30, 8,
		55, 0,
		80, 7,
		113, 43
	}
	
	for i=1, #chassisPoly1, 2 do
		v = chassisPoly1[i]
		v2 = chassisPoly1[i+1]

		print("X: " .. chassisPoly1[i] .. " Y: " .. chassisPoly1[i+1] )
		chassisPoly1[i] = v - 115/2
		chassisPoly1[i+1] = v2 - 68/2
	end
	
	local chassisPoly2 = {
		0, 49,
		4, 57,
		22, 54,
		34, 66,
		77, 66,
		93, 54,
		108, 57,
		114,43
	}
	
	for i=1, #chassisPoly2, 2 do
		v = chassisPoly2[i]
		v2 = chassisPoly2[i+1]

		print("X: " .. chassisPoly2[i] .. " Y: " .. chassisPoly2[i+1] )
		chassisPoly2[i] = v - 115/2
		chassisPoly2[i+1] = v2 - 68/2
	end
	
	local chassisWidth = 60
	local chassisHeight = 40
	local leftWheelSize = 16
	local rightWheelSize = 16
	
	o.bodies = {
		chassis = world:newBody( x, y, 0, 0),
		leftWheelAnchor = world:newBody( x-chassisWidth/1.5, y+leftWheelSize*2, 0, 0),
		rightWheelAnchor = world:newBody( x+chassisWidth/1.5, y+rightWheelSize*2, 0, 0),
		leftWheel = world:newBody( x-chassisWidth/1.5, y+leftWheelSize*2, 0, 0),
		rightWheel = world:newBody( x+chassisWidth/1.5, y+rightWheelSize*2, 0, 0),
	}
	
	o.shapes = {
		--chassis = o.bodies.chassis:newRectangleShape(0, 0, chassisWidth, chassisHeight),
		chassis = o.bodies.chassis:newPolygonShape(unpack(chassisPoly1)),
		chassis2 = o.bodies.chassis:newPolygonShape(unpack(chassisPoly2)),
		leftWheel = o.bodies.leftWheel:newCircleShape(0, 0, leftWheelSize/2),
		rightWheel = o.bodies.rightWheel:newCircleShape(0, 0, rightWheelSize/2),

		leftWheelAnchor = o.bodies.leftWheelAnchor:newCircleShape(0, 0, leftWheelSize),
		rightWheelAnchor = o.bodies.rightWheelAnchor:newCircleShape(0, 0, rightWheelSize)
	}
	
	o.joints = {
		rightSuspension = o.bodies.rightWheel:newDistanceJoint(o.bodies.chassis, 0, 0, chassisWidth/3, 0),
		leftSuspension = o.bodies.leftWheel:newDistanceJoint(o.bodies.chassis, 0, 0, -chassisWidth/3, 0),
	
		rightSuspensionRail = o.bodies.rightWheel:newPrismaticJoint(o.bodies.chassis, 0, 0, 0, 1),
		leftSuspensionRail = o.bodies.leftWheel:newPrismaticJoint(o.bodies.chassis, 0, 0, 0, 1),
			
		leftWheelAxel = o.bodies.leftWheel:newRevoluteJoint(o.bodies.leftWheelAnchor, 0, 0),
		rightWheelAxel = o.bodies.rightWheel:newRevoluteJoint(o.bodies.rightWheelAnchor, 0, 0),
		
		crossBar = o.bodies.leftWheel:newDistanceJoint(o.bodies.rightWheel, 0, 0)
	}
	
	o.images = {
		chassis = love.graphics.newImage("gfx/chassis.png"),
		wheel = love.graphics.newImage("gfx/wheel.png"),
	}
	
	--For Quick Ref
	local b, s, j = o.bodies, o.shapes, o.joints
	
	--Shape Settings
	s.chassis:setFriction(0)
	s.leftWheelAnchor:setFriction(1)
	s.rightWheelAnchor:setFriction(1)
	s.chassis:setRestitution(1)
	
	s.chassis:setCategory(3)
	s.chassis2:setCategory(3)
	s.leftWheel:setCategory(3) --May fuck up
	s.rightWheel:setCategory(3) -- Ditto
	
	s.leftWheel:setMask(3)
	s.rightWheel:setMask(3)
	s.leftWheelAnchor:setMask(3)
	s.rightWheelAnchor:setMask(3)
		
	--Jointing
	j.leftWheelAxel.joint:setMotorEnabled(true)
	j.rightWheelAxel.joint:setMotorEnabled(true)
	j.leftWheelAxel.joint:setMaxMotorTorque(1.4)
	j.rightWheelAxel.joint:setMaxMotorTorque(1.4)
	
	j.crossBar.joint:setLength( j.crossBar.joint:getLength() )
	j.crossBar.joint:setFrequency(3)
	j.crossBar.joint:setDamping(0.85)
	
	j.rightSuspension.joint:setLength( j.rightSuspension.joint:getLength() * 1.3 )
	j.leftSuspension.joint:setLength( j.leftSuspension.joint:getLength() * 1.3 )
	
	j.rightSuspension.joint:setFrequency(2.5)
	j.leftSuspension.joint:setFrequency(2.5)
	j.rightSuspension.joint:setDamping(0.2)
	j.leftSuspension.joint:setDamping(0.2)
	
	j.rightSuspensionRail.joint:setLimitsEnabled(true)
	j.leftSuspensionRail.joint:setLimitsEnabled(true)
	j.rightSuspensionRail.joint:setLimits(-80, 10)
	j.leftSuspensionRail.joint:setLimits(-80, 10)
	
	for i, body in pairs(o.bodies) do
		body:setMassFromShapes()
		--body:setDebugDraw(false)
	end
	
	setmetatable(o, Garage.Vehicle)
	return o
end

function Garage.Vehicle:update(dt)
end

function Garage.Vehicle:draw()
	--For Quick Ref
	local b, s, j, img = self.bodies, self.shapes, self.joints, self.images
	
	local carX, carY, carAng = b.chassis:getX(), b.chassis:getY(), b.chassis:getAngle()
	local carLWx, carLWy = b.leftWheelAnchor:getPosition()
	local carRWx, carRWy = b.rightWheelAnchor:getPosition()
	local susLeftAnc1, susRightAnc1 =  ( vector(-50, 15) ):rotate_inplace( carAng ) , ( vector(20, 20) ):rotate_inplace( carAng )
	local susLeftAnc2, susRightAnc2 = ( vector(-25, 20) ):rotate_inplace( carAng ), ( vector(50, 15) ):rotate_inplace( carAng )
		
	love.graphics.setLineWidth(6)
	love.graphics.setColor(50, 50, 50)
		
	love.graphics.line(carX + susLeftAnc1.x, carY + susLeftAnc1.y,  carLWx, carLWy ) -- Left Suspension / Left Bar
	love.graphics.line(carX + susLeftAnc2.x, carY + susLeftAnc2.y, carLWx, carLWy ) -- Left Suspension / Right Bar
		
	love.graphics.line(carX + susRightAnc1.x, carY + susRightAnc1.y, carRWx, carRWy)
	love.graphics.line(carX + susRightAnc2.x, carY + susRightAnc2.y, carRWx, carRWy)
	
	love.graphics.setColor(255,255,255)
	love.graphics.draw(img.chassis, carX, carY, carAng, 1, 1, img.chassis:getWidth()/2, img.chassis:getHeight()/2 )
	
	--love.graphics.circle("fill", carX + susLeftAnc1.x, carY + susLeftAnc1.y, 2)
	--love.graphics.circle("fill", carX + susRightAnc1.x, carY + susRightAnc1.y, 2)
	--love.graphics.circle("fill", carX + susLeftAnc2.x, carY + susLeftAnc2.y, 2)
	--love.graphics.circle("fill", carX + susRightAnc2.x, carY + susRightAnc2.y, 2)
	
	love.graphics.draw(img.wheel, b.leftWheelAnchor:getX(), b.leftWheelAnchor:getY(), b.leftWheelAnchor:getAngle(), 1, 1, img.wheel:getWidth()/2, img.wheel:getHeight()/2 )
	love.graphics.draw(img.wheel, b.rightWheelAnchor:getX(), b.rightWheelAnchor:getY(), b.rightWheelAnchor:getAngle(), 1, 1, img.wheel:getWidth()/2, img.wheel:getHeight()/2 )
	
	love.graphics.setLineWidth(1)
	--fuzz.defaultDraw()
end