--Gymnast Class
print("Loaded Gymnast class")
Gymnast = {}
Gymnast.__index = Gymnast
Gymnast.list = {}

--Tweakable Settings for Gymnast
Arm_MaxSpeed = 2
UpperLeg_MaxSpeed = 2 
LowerLeg_MaxSpeed = 2

function Gymnast.create(x, y, angle)
	local o = {}
	
	o.bodies = {
		body = fizz.newBody(world, x, y, 0, 0),
		head = fizz.newBody(world, x+17.5, y, 0, 0),
		arm = fizz.newBody(world, x+25, y, 0, 0),
		upperleg = fizz.newBody(world, x-30, y, 0, 0),
		lowerleg = fizz.newBody(world, x-50, y, 0, 0)
	}
	
	o.shapes = {
		body = 	o.bodies.body:newRectangleShape(0, 0, 40, 6, 0),
		head = 	o.bodies.head:newCircleShape(0, 0, 6),
		arm = o.bodies.arm:newRectangleShape(0, 0, 25, 3, 0),
		upperleg = o.bodies.upperleg:newRectangleShape(0, 0, 25, 6, 0),
		lowerleg = o.bodies.lowerleg:newRectangleShape(0, 0, 15, 6, 0)
	}
	
	o.joints = {
		grab1 = false,
		grab2 = false,
		grabcannon = false,
		head = o.bodies.head:newRevoluteJoint(o.bodies.body, 0, 0),
		arm = o.bodies.arm:newRevoluteJoint(o.bodies.body, -15, 0),
		upperleg = o.bodies.upperleg:newRevoluteJoint(o.bodies.body, 25/2, 0),
		lowerleg = o.bodies.lowerleg:newRevoluteJoint(o.bodies.upperleg, 7.5, 0)
	}
	
	--Set all bodies mass automatically
	for i, limb in pairs(o.bodies) do
		limb:setMassFromShapes()
	end
	
	--Stop leg from sleeping
	o.bodies.upperleg.body:setAllowSleeping(false)
	o.bodies.lowerleg.body:setAllowSleeping(false)
	
	--Set Collision Filtering (Head messes with arm)
	o.shapes.body:setCategory(1)
	o.shapes.head:setMask(1)
	
	--Set Joints
		--Arm
		o.joints.arm.joint:setMotorEnabled(true)
		o.joints.arm.joint:setMaxMotorTorque(1)
		
		--UpperLeg
		o.joints.upperleg.joint:setLimitsEnabled(true)
		o.joints.upperleg.joint:setLimits( math.rad(-30), math.rad(120) )
		o.joints.upperleg.joint:setMotorEnabled(true)
		o.joints.upperleg.joint:setMaxMotorTorque(2)
	
		--Lowerleg
		o.joints.lowerleg.joint:setLimitsEnabled(true)
		o.joints.lowerleg.joint:setLimits( math.rad(-90), math.rad(0) )
		o.joints.lowerleg.joint:setMotorEnabled(true)
		o.joints.lowerleg.joint:setMaxMotorTorque(200)
	
	
	setmetatable(o, Gymnast)
	table.insert(Gymnast.list, o)
	
	return o
end

function Gymnast:isGrabbing()
	if self.mode == "grab" then
		return true
	end
	return false
end

function Gymnast:isCannoning()
	if self.mode == "cannon" then
		return true
	end
	return false
end

function Gymnast:update()
	if self.mode == "grab" then
		if self:isCannoning() then
			self:detach()
		end
		gymguy.joints.upperleg.joint:setMotorSpeed( math.rad( love.joystick.getAxis(0, 0) * 280) )
		gymguy.joints.lowerleg.joint:setMotorSpeed( math.rad( love.joystick.getAxis(0, 0) * 280) )
	elseif self.mode == "cannon" then
		if self:isGrabbing() then
			self:detach()
		end
		if love.joystick.getAxis(0,2) < -0.2 then --With R trigger
			self.joints.arm.joint:setMotorSpeed(-100)
			self.joints.upperleg.joint:setMotorSpeed( love.joystick.getAxis(0,2) * -UpperLeg_MaxSpeed )
			self.joints.lowerleg.joint:setMotorSpeed( love.joystick.getAxis(0,2) * -LowerLeg_MaxSpeed )
			self.bodies.body:applyTorque( love.joystick.getAxis(0, 0) * math.abs(love.joystick.getAxis(0,2) / 5) * 1.5 )
		else
			self.joints.arm.joint:setMotorSpeed(100)
			self.joints.upperleg.joint:setMotorSpeed( -460 )
			self.joints.lowerleg.joint:setMotorSpeed( 280 )
			self.bodies.body:applyTorque( love.joystick.getAxis(0, 0) / 2 )
			self.bodies.body:setAngularVelocity( self.bodies.body:getAngularVelocity() / 1.05 )
		end
	end
end

function Gymnast:detach()
	if self.joints.grab1 and self.joints.grab2 then
		self.joints.grab1:destroy()
		self.joints.grab2:destroy()
		self.joints.grab1 = false
		self.joints.grab2 = false
		print("Deleted grab joints")
	end
	
	if self.joints.grabcannon then
		self.joints.grabcannon:destroy()
		self.joints.grabcannon = false
		print("Deleted cannon joints")
	end
end

--[[function Gymnast:setMode(mode)
	if mode == "grab" or mode == "cannon" then
		self.mode = mode
		print("Set mode to " .. mode)
		return true
	end
	return false
end]]--

function Gymnast:getMode()
	return self.mode
end

function Gymnast:destroy()
end

function Gymnast:getNearestBody()
end

function Gymnast:getAttachablePoint(body)
end

--Gymnastic Mode Functions (For Joints)
function Gymnast:setGrab(target)
	if self:isCannoning() then
		self:detach()
	end
	
	--Arm
	self.joints.arm.joint:setMotorEnabled(true)
	self.joints.arm.joint:setMaxMotorTorque(1)
	
	--UpperLeg
	self.joints.upperleg.joint:setLimitsEnabled(true)
	self.joints.upperleg.joint:setLimits( math.rad(-30), math.rad(120) )
	self.joints.upperleg.joint:setMotorEnabled(true)
	self.joints.upperleg.joint:setMaxMotorTorque(2)

	--Lowerleg
	self.joints.lowerleg.joint:setLimitsEnabled(true)
	self.joints.lowerleg.joint:setLimits( math.rad(-90), math.rad(0) )
	self.joints.lowerleg.joint:setMotorEnabled(true)
	self.joints.lowerleg.joint:setMaxMotorTorque(200)
	
	--Target Joint
	hand = vector(12.5, 0)
	hand:rotate_inplace(gymguy.bodies.arm.body:getAngle())
	
	self.joints.grab1 = self.bodies.arm:newDistanceJoint(target, hand.x, hand.y+3)
	self.joints.grab2 = self.bodies.arm:newDistanceJoint(target, hand.x, hand.y-3)
	
	self.joints.grab1.joint:setLength(7)
	self.joints.grab2.joint:setLength(7)
	
	self.mode = "grab"
end

function Gymnast:setCannon()
	if self:isGrabbing() then
		self:detach()
	end
	
	--Arm v0.2
	self.joints.arm.joint:setLimitsEnabled(true)
	self.joints.arm.joint:setLimits(math.rad(-135), math.rad(0))
	self.joints.arm.joint:setMaxMotorTorque(1)
	self.joints.arm.joint:setMotorEnabled(true)
	self.joints.arm.joint:setMotorSpeed(-Arm_MaxSpeed)
	
	--Arm
	--self.joints.arm.joint:setLimitsEnabled(false)
	
	--UpperLeg
	self.joints.upperleg.joint:setLimitsEnabled(true)
	self.joints.upperleg.joint:setLimits( math.rad(-0), math.rad(120) )
	self.joints.upperleg.joint:setMaxMotorTorque(2)
	self.joints.upperleg.joint:setMotorSpeed( UpperLeg_MaxSpeed )
	
	--LowerLeg
	--self.joints.lowerleg.joint:setMotorSpeed( -280 )
	self.joints.lowerleg.joint:setMaxMotorTorque(2)
	self.joints.lowerleg.joint:setMotorSpeed( -LowerLeg_MaxSpeed)
	
	--Grab Leg
	if not self.joints.grabcannon then
		--self.joints.grabcannon = self.bodies.arm:newDistanceJoint(self.bodies.lowerleg)
		--self.joints.grabcannon.joint:setFrequency(5)
		--self.joints.grabcannon.joint:setLength( 15 )
	end
	self.mode = "cannon"
end